require 'helpers'

local iv = require 'nibui.InterpolatedValue'
local Easing = require 'nibui.Easing'

local Scene = require 'scene'

local TowerScene = Scene:new()

local Player = require 'player'
PrepareHeroesScene = PrepareHeroesScene or require 'scenes.prepare_heroes'
GameOverScene = GameOverScene or require 'scenes.game_over'
CombatScene = CombatScene or require 'scenes.combat'

function TowerScene:new()
    local tower = new(TowerScene, {
        clouds = {},
        ropes = {
            {
                start = {180, 252},
                length = 3
            },
            {
                start = {94, 275},
                length = 10
            },
            {
                start = {150, 376},
                length = 2
            },
            {
                start = {156, 372},
                length = 8
            },
            {
                start = {209, 364},
                length = 1
            },
            {
                start = {213, 345},
                length = 3
            }
        },

        -- x, y, source_x, source_y, w, h, palette
        background = {320, 0, 320, 960, 1},
        background_tilt = 31,
        angle = iv:new(0),
        y = 0,
        x = 0,

        mouse_waving_amplitude = 10,
        waving_amplitude = 2,
        waving_frequency = 4,
    })

    local big_clouds = {
        {208, 32, 73, 7, 1},
        {208, 48, 79, 16, 1},
        {208, 80, 53, 11, 1},
    }

    local small_clouds = {
        {208, 43, 20, 5, 1},
        {208, 64, 22, 5, 1},
        {240, 64, 18, 4, 1},
    }

    local far_clouds = {
        {240, 24, 8, 2, 1},
        {240, 28, 5, 2, 1},
        {208, 16, 25, 4, 1},
    }

    for speed=60,2,-10 do
        local clouds, amount
        
        if speed > 45 then
            clouds = big_clouds
            amount = 32
            speed = 64
        elseif speed > 30 then
            clouds = small_clouds
            amount = 128
            speed = 16
        else
            clouds = far_clouds
            amount = 128
            speed = 8
        end

        for i=1, amount do
            push(tower.clouds, {
                x = math.random(0, 640),
                y = math.random(100, 500),
                vx = speed+math.random()*10,
                vy = 0,
                sprite = clouds[math.random(1, #clouds)]
            })
        end
    end

    return tower
end

function TowerScene:angle2y(angle)
    local angle_ratio = self.background[4]/180

    return angle_ratio*(angle-self.background_tilt)
end

function TowerScene:wave()
    local wave = math.sin(clock()+math.sin(clock())*self.waving_frequency)*self.waving_amplitude
    local x, y = mouse_position()

    return (120-y)/240*self.mouse_waving_amplitude+wave
end

function TowerScene:draw()
    clip(0, 0, 320, 240)
    
    local y = self:angle2y(self.angle.value)

    self.y = y

    clear(12+16)

    for _, cloud in ipairs(self.clouds) do
        custom_sprite(cloud.x-160, y/2-100+cloud.y, unwrap(cloud.sprite))
    end

    custom_sprite(0, y, unwrap(self.background))

    sprite(122, y+348, 13+(math.floor(clock()*8)%7), 0, 1)

    for _, rope in ipairs(self.ropes) do
        local center = {160, 120}

        local dx, dy = rope.start[1]-center[1], (y+rope.start[2])-center[2]

        dx /= rope.length
        dy /= rope.length

        line(
            rope.start[1], rope.start[2]+y,
            dx+rope.start[1], dy+rope.start[2]+y,
            6+16
        )
    end

    -- super
    Scene.draw(self)
end

function TowerScene:update(dt)
    self.angle:update(dt)

    -- super
    Scene.update(self, dt)

    for _, cloud in ipairs(self.clouds) do
        cloud.x -= cloud.vx*dt
        cloud.y -= cloud.vy*dt

        cloud.x = cloud.x%640
    end
end

function TowerScene:opening()
    set_timeout(function()
        self.waving_amplitude = 0
        self.mouse_waving_amplitude = 16
        self.angle:set(0, 1, Easing.InOutCubic)

        set_timeout(function()
            self:finish_scene()
            self:start_scene(GameOverScene:new())
        end, 0)
    end, 0.1)
end

function TowerScene:configuration(p1, p2)
    set_timeout(function()
        self.waving_amplitude = 0.5
        self.mouse_waving_amplitude = 0
        self.angle:set(-61, 1, Easing.InOutCubic)

        set_timeout(function()
            local p1, p2 = p1 or Player:new(11, true), p2 or Player:new()
            self:finish_scene()
            self:start_scene(PrepareHeroesScene:new(p1, p2))
        end, 1)
    end, 0.1)
end

function TowerScene:combat(p1, p2)
    set_timeout(function()
        self:start_scene(CombatScene:new(p1, p2))

        self.waving_amplitude = 0.5
        self.mouse_waving_amplitude = 2
        self.angle:set(-100, 1, Easing.InOutCubic)

        set_timeout(function()
            self:finish_scene(-1)
        end, 1)
    end, 0.1)
end

return TowerScene
